#!/usr/bin/env node
import { command, cli } from 'cleye';
import _ from 'lodash';
import * as p from '@clack/prompts';
import { dim, red, green, blue, lightGreen, lightYellow, bgYellow, black } from 'kolorist';
import dedent from 'dedent';
import { promisify } from 'node:util';
import { exec, spawn } from 'node:child_process';
import { access } from 'node:fs';
import path from 'path';
import fs from 'node:fs/promises';
import os from 'os';
import ini from 'ini';

var name = "commitizen-cli";
var version = "3.1.2";
var author = {
	name: "Ando",
	url: "https://github.com/iamando",
	email: "iamando@tuta.io"
};

const PACKAGE_NAME = name;
const VERSION = version;
const AUTHOR = author;
const CANCELED_OP_MSG = "Operation Canceled";
const commitKeys = Object.freeze({
  type: "<type>",
  ticket: "<ticket>",
  commit: "<commit>"
});

const type = () => p.select({
  message: "Choose commit type",
  initialValue: "feat",
  options: [
    {
      label: "Feat",
      value: "feat",
      hint: "For new features (e.g., feat: implement --cache-to feature to export cache)"
    },
    {
      label: "Fix",
      value: "fix",
      hint: "For improvements and bugfixes that do not introduce a feature (e.g., fix: improve error message)"
    },
    {
      label: "Hotfix",
      value: "hotfix",
      hint: "For improvements and bugfixes that do not introduce a feature, directly in prodution mode (e.g., fix: improve success message)"
    },
    {
      label: "Chore",
      value: "chore",
      hint: "General things that should be excluded (e.g., chore: clean up X)"
    },
    {
      label: "Design",
      value: "design",
      hint: "For design changes only (e.g., design: use rounded button)"
    },
    {
      label: "Experiment",
      value: "experiment",
      hint: "General things that should be in experiment (e.g., experiment: implement new query system in X)"
    },
    {
      label: "Docs",
      value: "docs",
      hint: "For documentation changes only (e.g., docs: fix typo in X)"
    },
    {
      label: "Refact",
      value: "refact",
      hint: "General things that should be restructured but not changing the orginal functionality (e.g., refact: move X to new file utils)"
    },
    {
      label: "CI",
      value: "ci",
      hint: "For internal CI specific changes (e.g., ci: enable X for tests)"
    },
    {
      label: "Infra",
      value: "infra",
      hint: "For infrastructure changes (e.g., infra: Enable cloudfront for X)"
    },
    {
      label: "Test",
      value: "test",
      hint: "For changes to tests only (e.g., test: check if X does Y)"
    }
  ]
});
const commit = () => p.text({
  message: "Insert commit message",
  validate: (value) => {
    if (_.isEmpty(value))
      return "Commit message required";
  }
});
const ticket = async ({ flags }) => {
  if (flags.minified)
    return;
  const hasTicket = await p.confirm({
    message: "Has ticket ?",
    initialValue: false
  });
  if (!hasTicket)
    return;
  await p.text({
    message: "Insert ticket name",
    validate: (value) => {
      if (_.isEmpty(value))
        return "Ticket name required";
    }
  });
};

const log = ({
  type,
  msg,
  isConsole = true,
  newLine = true
}) => {
  switch (type) {
    case "info":
      return isConsole ? console.info(`${newLine ? "\n" : ""} ${blue("\u2754")} ${msg}`) : `${newLine ? "\n" : ""} ${blue("\u2754")} ${msg}`;
    case "success":
      return isConsole ? console.log(`${newLine ? "\n" : ""} ${green("\u2714")} ${msg}`) : `${newLine ? "\n" : ""} ${green("\u2714")} ${msg}`;
    case "error":
      return isConsole ? console.error(` ${newLine ? "\n" : ""} ${red("\u274C")} ${msg}`) : `${newLine ? "\n" : ""} ${red("\u274C")} ${msg}`;
    default:
      return isConsole ? console.log(`${newLine ? "\n" : ""} ${dim("\u2755")} ${msg}`) : `${newLine ? "\n" : ""} ${dim("\u2755")} ${msg}`;
  }
};

class CliError extends Error {
}
const indent = " ".repeat(4);
const handleCliError = (error) => {
  if (error instanceof Error && !(error instanceof CliError)) {
    if (error.stack) {
      log({
        type: "error",
        msg: error.stack.split("\n").slice(1).join("\n")
      });
    }
    log({
      type: "error",
      msg: `
${indent}${dim(`${PACKAGE_NAME} v${VERSION}`)}`
    });
    log({
      type: "error",
      msg: `
${indent}Please open a Bug report with the information above:`
    });
    log({
      type: "error",
      msg: `${indent}https://github.com/${AUTHOR.name}/${PACKAGE_NAME}/issues/new`
    });
  }
};

const execa = promisify(exec);
const gitStatus = async () => {
  const { stdout: stdoutStatus, stderr: stderrStatus } = await execa(
    "git status"
  );
  return { stdoutStatus, stderrStatus };
};
const gitAdd = async () => {
  const { stdout: stdoutAdd, stderr: stderrAdd } = await execa("git add .");
  return { stdoutAdd, stderrAdd };
};
const gitCommit = async ({ commit }) => {
  const { stdout: stdoutCommit, stderr: stderrCommit } = await execa(
    `git commit -m "${commit}"`
  );
  return { stdoutCommit, stderrCommit };
};
const isTreeClean = async () => {
  const { stdoutStatus, stderrStatus } = await gitStatus();
  if (stderrStatus)
    throw new CliError(`An error occured: ${stderrStatus}`);
  if (stdoutStatus.includes("nothing to commit, working tree clean")) {
    p.outro(lightGreen("Nothing to commit, working tree clean \u{1F9F9}"));
    process.exit(1);
  }
};
const isGitRepository = () => {
  const dir = path.resolve(".git");
  access(dir, (err) => {
    if (err && err.code === "ENOENT") {
      p.outro(lightYellow("Not a git repository \u{1F622}"));
      process.exit(1);
    }
  });
};

const { hasOwnProperty } = Object.prototype;
const hasOwn = (object, key) => hasOwnProperty.call(object, key);
const configParsers = {
  PATTERN(pattern) {
    if (!pattern)
      return "";
    return pattern;
  }
};
const configPath = path.join(os.homedir(), ".commitizen-cli");
const fileExists = (filePath) => fs.lstat(filePath).then(
  () => true,
  () => false
);
const readConfigFile = async () => {
  const configExists = await fileExists(configPath);
  if (!configExists) {
    return /* @__PURE__ */ Object.create(null);
  }
  const configString = await fs.readFile(configPath, "utf8");
  return ini.parse(configString);
};
const getConfig = async (cliConfig) => {
  const config = await readConfigFile();
  const parsedConfig = {};
  for (const key of Object.keys(configParsers)) {
    const parser = configParsers[key];
    const value = cliConfig?.[key] ?? config[key];
    parsedConfig[key] = parser(value);
  }
  return parsedConfig;
};
const setConfigs = async (keyValues) => {
  const config = await readConfigFile();
  for (const [key, value] of keyValues) {
    if (!hasOwn(configParsers, key)) {
      throw new CliError(`Invalid config property: ${key}`);
    }
    const parsed = configParsers[key](value);
    config[key] = parsed;
  }
  await fs.writeFile(configPath, ini.stringify(config), "utf8");
};

const pattern = async ({
  type,
  commit,
  ticket
}) => {
  let defaultPattern;
  const { PATTERN: configPattern } = await getConfig();
  if (_.isEmpty(configPattern)) {
    ticket ? defaultPattern = `${commitKeys.type}(${commitKeys.ticket}): ${commitKeys.commit}` : defaultPattern = `${commitKeys.type}: ${commitKeys.commit}`;
  } else {
    defaultPattern = configPattern;
  }
  const hasPattern = await p.confirm({
    message: "Has specific pattern ?",
    initialValue: false
  });
  if (p.isCancel(hasPattern)) {
    p.cancel(CANCELED_OP_MSG);
    process.exit(0);
  }
  if (!hasPattern)
    return defaultPattern;
  const pattern2 = await p.text({
    message: "Insert specific pattern",
    placeholder: `Example: ${commitKeys.type}(${commitKeys.ticket}): ${commitKeys.commit}`,
    defaultValue: defaultPattern,
    validate: (value) => {
      if (_.isEmpty(value))
        return "Specific pattern required";
      if (!hasPatternKeys({ type, commit, ticket, pattern: value }))
        return `Pattern key not invalid (ex: ${commitKeys.type}, ${commitKeys.ticket}, ${commitKeys.commit})`;
    }
  });
  if (p.isCancel(pattern2)) {
    p.cancel(CANCELED_OP_MSG);
    process.exit(0);
  }
  const savePatterConfig = await p.confirm({
    message: "Save pattern to config ?",
    initialValue: true
  });
  if (p.isCancel(savePatterConfig)) {
    p.cancel(CANCELED_OP_MSG);
    process.exit(0);
  }
  setConfigs([["PATTERN", pattern2]]);
};
const hasPatternKeys = ({
  type,
  commit,
  ticket,
  pattern: pattern2
}) => {
  const patternKeys = [];
  if (!_.isEmpty(type))
    patternKeys.push(commitKeys.type);
  if (!_.isEmpty(commit))
    patternKeys.push(commitKeys.commit);
  if (!_.isEmpty(ticket))
    patternKeys.push(commitKeys.ticket);
  const containKeys = patternKeys.every((key) => pattern2.includes(key));
  return containKeys;
};
const replace = ({
  pattern: pattern2,
  values
}) => {
  for (const value in values) {
    pattern2 = pattern2?.replace(
      new RegExp(value, "g"),
      values[value]
    );
  }
  return pattern2;
};
const transform = async ({
  type,
  commit,
  ticket,
  pattern: pattern2
}) => {
  const patternKeys = {
    [commitKeys.type]: type,
    [commitKeys.ticket]: ticket ?? "",
    [commitKeys.commit]: commit
  };
  const message = replace({ pattern: pattern2, values: patternKeys });
  return message;
};

const commiter = async ({ flags }) => {
  p.intro(bgYellow(black("Commitizen CLI")));
  isGitRepository();
  await isTreeClean();
  const values = await p.group(
    {
      type: () => type(),
      commit: () => commit(),
      ticket: () => ticket({ flags })
    },
    {
      onCancel: () => {
        p.cancel(CANCELED_OP_MSG);
        process.exit(0);
      }
    }
  );
  try {
    let commit2;
    if (flags.minified) {
      commit2 = `${values.type}: ${values.commit}`;
    } else {
      const blueprint = await pattern({
        type: values.type,
        commit: values.commit,
        ticket: values.ticket
      });
      if (_.isEmpty(blueprint))
        return;
      commit2 = await transform({
        type: values.type,
        commit: values.commit,
        ticket: values.ticket,
        pattern: blueprint
      });
    }
    const { stdoutStatus, stderrStatus } = await gitStatus();
    if (stderrStatus)
      throw new CliError(`An error occured: ${stderrStatus}`);
    if (stdoutStatus.includes("no changes added to commit") || stdoutStatus.includes(
      "nothing added to commit but untracked files present"
    )) {
      const type2 = stdoutStatus.includes("no changes added to commit") ? "modified" : "untracked";
      const addStagedFiles = await p.confirm({
        message: `No changes added to commit, would you like to add ${type2} files ?`,
        initialValue: true
      });
      if (addStagedFiles) {
        const { stderrAdd } = await gitAdd();
        if (stderrAdd)
          throw new CliError(`An error occured: ${stderrAdd}`);
        const { stderrCommit: stderrCommit2 } = await gitCommit({ commit: commit2.trim() });
        if (stderrCommit2)
          throw new CliError(`An error occured: ${stderrCommit2}`);
        p.outro(dedent`
        You're all set 🎉

        use "git push" to publish your local commits
      `);
      }
      return;
    }
    const { stderrCommit } = await gitCommit({ commit: commit2.trim() });
    if (stderrCommit)
      throw new CliError(`An error occured: ${stderrCommit}`);
    p.outro(dedent`
    You're all set 🎉

    use "git push" to publish your local commits
  `);
  } catch (err) {
    log({ type: "error", msg: err.message });
    handleCliError(err);
    process.exit(1);
  }
};

var update = command(
  {
    name: "update",
    description: "Update Commitizen CLI to the latest version"
  },
  async () => {
    const cmd = "npm update -g commitizen-cli";
    console.log(dim(`Running: ${cmd}`));
    const update = spawn(cmd, {
      stdio: "inherit",
      shell: process.env.SHELL || true
    });
    update.stderr?.on("error", () => {
      throw new CliError("An error occured on updating package");
    });
    update.on("close", (code) => {
      if (code !== 0) {
        console.log(`update process exited with code ${code}`);
      }
    });
  }
);

var config = command(
  {
    name: "config",
    parameters: ["[mode]", "[key=value...]"],
    description: "Get or Set Commitizen CLI config"
  },
  (argv) => {
    (async () => {
      const { mode, keyValue: keyValues } = argv._;
      if (!keyValues.length) {
        console.error('Error: Missing required parameter "key=value"\n');
        argv.showHelp();
        return process.exit(1);
      }
      if (mode === "get") {
        const config = await getConfig();
        for (const key of keyValues) {
          if (hasOwn(config, key)) {
            console.log(`${key}=${config[key]}`);
          }
        }
        return;
      }
      if (mode === "set") {
        await setConfigs(
          keyValues.map(
            (keyValue) => keyValue.split("=")
          )
        );
        return;
      }
      throw new CliError(`Invalid mode: ${mode}`);
    })().catch((err) => {
      log({ type: "error", msg: err.message });
      handleCliError(err);
      process.exit(1);
    });
  }
);

var clean = command(
  {
    name: "clean",
    description: "Clean Commitizen CLI configuration"
  },
  async () => {
    const cmd = "commitizen clean";
    console.log(dim(`Running: ${cmd}`));
    const configPath = path.join(os.homedir(), ".commitizen-cli");
    fs.writeFile(configPath, "").then(() => {
      log({ type: "success", msg: "Configuration cleaned successfully" });
    }).catch(() => {
      throw new CliError("An error occured while cleaning configuration");
    });
  }
);

cli(
  {
    name: PACKAGE_NAME,
    version: VERSION,
    commands: [update, config, clean],
    flags: {
      minified: {
        type: Boolean,
        description: "Use the minified commit structure"
      }
    }
  },
  ({ flags }) => {
    commiter({ flags }).catch((err) => {
      log({ type: "error", msg: err.message });
      handleCliError(err);
      process.exit(1);
    });
  }
);
